# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTIBILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

#bl_info = {
#    "name" : "Hologram Camera",
#    "author" : "Endre Sipos",
#    "description" : "Add-on for placing and parametrizing a camera for rendering images needed for a hologram generation. Images shall be rendered with the Render Animation command.",
#    "blender" : (2, 82, 0),
#    "version" : (1, 3, 0),
#    "location" : "View3D > Add > Hologram Camera",
#    "warning" : "",
#    "category" : "Hologram Camera"
#}

import bpy
import time
import math

from bpy.types import (
        Menu,
        Operator
        )
        
from bpy.props import (
        FloatProperty,
        IntProperty,
        StringProperty, 
        IntProperty, 
        BoolProperty,
        EnumProperty
        )        

CAM_DISTANCE_COEFFICIENT = 2.07
CAM_TRACK_COEFFICIENT = 1.28
MM_TO_M = 0.001

FRAME_NAME = "Hologram frame"
CAMERA_NAME = "Hologram camera"


# ------------------------------------------------------------

class Camera:

#Camera position   
    def camera_distance(self, width, height):
        if width > height:
            return width * CAM_DISTANCE_COEFFICIENT
        return height * CAM_DISTANCE_COEFFICIENT; 
    

    def camera_track(self, cameraDistance):
        return CAM_TRACK_COEFFICIENT * cameraDistance 

# Image
    def vertical_pixel_count(self, hologramHeight, hogelSize):
        return hologramHeight / hogelSize

    def horizontal_pixel_count(self, cameraTrack, hologramWidth, hogelSize):
        pass;


# Rectangle:

    def get_rectangle_vertices(self, width, height):
        newpoints = []

        x = width / 2
        y = height / 2

        newpoints.append([-x, 0.0, y, 1.0])
        newpoints.append([x, 0.0, y, 1.0])
        newpoints.append([x, 0.0, -y, 1.0])
        newpoints.append([-x, 0.0, -y, 1.0])
        newpoints.append([-x, 0.0, y, 1.0])

        return newpoints
# Samples:
    def get_sample_count(self, props):
        return props.number_of_samples;
        

# Objects

    def draw_rectangle(self, widthBlender, heightBlender):
        verts = self.get_rectangle_vertices(widthBlender, heightBlender)
        vertsBulked = []
        for point in verts:
            vertsBulked += point

        dataCurve = bpy.data.curves.new(FRAME_NAME, type='CURVE')  
        dataCurve.dimensions = '3D'        

        newSpline = dataCurve.splines.new(type='POLY')              
        newSpline.points.add(4)
        newSpline.points.foreach_set('co', vertsBulked)
        newSpline.use_endpoint_u = False 
    
         #create Object
        curveOB = bpy.data.objects.new(FRAME_NAME, dataCurve)

        # attach
        bpy.context.collection.objects.link(curveOB)
        curveOB.hide_render = True
        curveOB.select_set(True)

    
    def add_camera(self, cameraDistance, cameraTrack, orthoScale):

        camera = bpy.data.cameras.new(CAMERA_NAME) 
        camera.type = 'ORTHO'
        camera.ortho_scale = orthoScale
        camera.sensor_fit = 'VERTICAL'
        camera.dof.use_dof = True
        camera.dof.focus_distance = cameraDistance * MM_TO_M  
    
         #create Object
        cameraOB = bpy.data.objects.new(CAMERA_NAME, camera)
        cameraOB.location = ((-cameraTrack * MM_TO_M) / 2, -cameraDistance * MM_TO_M, 0)
        cameraOB.rotation_euler = (math.pi / 2, 0, 0)

        #attach
        bpy.context.collection.objects.link(cameraOB)
        #cameraOB.select_set(True)

        return(cameraOB)

    #movement
    def animate_camera(self, scene, cameraOB, start, end, cameraDistance, samples):
        pass
        

    # Output
    def adjust_output(self, resolutionX, resolutionY):
        bpy.context.scene.render.resolution_x = resolutionX
        bpy.context.scene.render.resolution_y = resolutionY



    def writeOutput(self, name, sampleCount, hologramWidth, hologramHeight, hogelSize, cameraType, dir):
        with open(dir + name + ".xml", "w", encoding="utf-8") as file:
            file.write('<?xml version="1.0" encoding="UTF-8"?>')
            file.write('<HologramCamera>')
            file.write('<CameraType>')
            file.write(cameraType)
            file.write('</CameraType>')
            file.write('<SamplesPerAxis>')
            file.write(str(sampleCount))
            file.write('</SamplesPerAxis>')
            file.write('<HologramWidth>')
            file.write(str(hologramWidth))
            file.write('</HologramWidth>')
            file.write('<HologramHeight>')
            file.write(str(hologramHeight))
            file.write('</HologramHeight>')
            file.write('<HogelSize>')
            file.write("{:.2f}".format(hogelSize))
            file.write('</HogelSize>')
            file.write('</HologramCamera>')
            file.close
            
    def SetupCameraObject(self, props, context, xml_dir):
        sampleCount = self.get_sample_count(props)
        hologramWidth = props.size_x
        hologramHeight = props.size_y    
        hogelSize = props.hogel_size
    
        heightBlender = hologramHeight * MM_TO_M
        widthBlender = hologramWidth * MM_TO_M
        
        cameraDistance = self.camera_distance(hologramWidth, hologramHeight)
        cameraTrack = self.camera_track(cameraDistance)
    
        resolutionY = self.vertical_pixel_count(hologramHeight, hogelSize)
        resolutionX = self.horizontal_pixel_count(cameraTrack, hologramWidth, hogelSize)
    
        self.draw_rectangle(widthBlender, heightBlender)    
    
        self.adjust_output(resolutionX, resolutionY)
    

        cameraOB = self.add_camera(cameraDistance, cameraTrack, heightBlender)
        
        self.animate_camera(context.scene, cameraOB, (-cameraTrack * MM_TO_M) / 2, (cameraTrack * MM_TO_M) / 2, cameraDistance * MM_TO_M, sampleCount) 
        context.scene.camera = cameraOB    
        
        self.writeOutput(cameraOB.name, sampleCount, hologramWidth, hologramHeight, hogelSize, props.camera_type, xml_dir)
    

class HPO_Simple_Camera(Camera):
    
    def horizontal_pixel_count(self, cameraTrack, hologramWidth, hogelSize):
        return (cameraTrack + hologramWidth) / hogelSize
    
    def animate_camera(self, scene, cameraOB, start, end, cameraDistance, samples):
        bpy.context.scene.frame_start = 1
        bpy.context.scene.frame_end = samples
    

        cameraOB.location[0] = start
        cameraOB.keyframe_insert(data_path="location", index = 0, frame=1)

        cameraOB.location[0] = end
        cameraOB.keyframe_insert(data_path="location", index = 0, frame=samples)
    
        for fc in cameraOB.animation_data.action.fcurves:
            fc.extrapolation = 'LINEAR'


class HPO_Hybrid_Fixed_Camera(Camera):
    
    def horizontal_pixel_count(self, cameraTrack, hologramWidth, hogelSize):
        return hologramWidth / hogelSize
    
    def animate_camera(self, scene, cameraOB, start, end, cameraDistance, samples):
        bpy.context.scene.frame_start = 1
        bpy.context.scene.frame_end = samples
    
        step = (end - start) / (samples - 1)
    
    
        for i in range(samples):
            locationX = start + (i * step)
        
            trackAngle = math.pi / 2

            if locationX != 0:
                trackAngle = math.atan2(cameraDistance, locationX)

            angle = (math.pi / 2) - trackAngle
            frame = i + 1 


            cameraOB.location[0] = locationX
            cameraOB.rotation_euler[2] = angle
            cameraOB.keyframe_insert(data_path="location", index = 0, frame = i + 1)
            cameraOB.keyframe_insert(data_path="rotation_euler", index = 2, frame = i + 1)
            
class FP_Hybrid_Fixed_Camera(Camera):
    
    def get_sample_count(self, props):
        return props.number_of_samples_FP;
    
    def horizontal_pixel_count(self, cameraTrack, hologramWidth, hogelSize):
        return hologramWidth / hogelSize
    
    def animate_camera(self, scene, cameraOB, start, end, cameraDistance, samples):
        frameCount = samples * samples

        bpy.context.scene.frame_start = 1
        bpy.context.scene.frame_end = frameCount

        frames = range(1, frameCount)
    
        step = (end - start) / (samples - 1)
        
        locationsZ = [0] * samples
        locationsX = [0] * samples
        anglesX = [0] * samples
        anglesZ = [0] * samples
        
        for s in range(samples):
            location = start + (s * step)

            locationsX[s] = location
            locationsZ[s] = -location
            
            trackAngle = math.pi / 2

            if location != 0:
                trackAngle = math.atan2(cameraDistance, location)
        
            anglesX[s] = math.pi / 2 - trackAngle
            anglesZ[s] = math.pi - trackAngle


        locationsXAll = locationsX * samples
        anglesXAll = anglesX * samples

        locationsZAll = [0] * frameCount;    
        anglesZAll = [0] * frameCount;    
        
        for t in range(samples):
            startIndex = samples * t
            endIndex = startIndex+samples-1
            locationsZAll[startIndex:endIndex] = [locationsZ[t]] * samples
            anglesZAll[startIndex:endIndex] = [anglesZ[t]] * samples
    

        a = bpy.data.actions.new("Move")
        
        fc = a.fcurves.new("location", index=2, action_group="locVertical")
        fc.keyframe_points.add(count=len(frames))
        fc.keyframe_points.foreach_set("co", [x for co in zip(frames, locationsZAll) for x in co])
        fc.update()
        fc.convert_to_keyframes(1, frameCount)
        
        fc = a.fcurves.new("rotation_euler", index=0, action_group="degVertical")
        fc.keyframe_points.add(count=len(frames))
        fc.keyframe_points.foreach_set("co", [x for co in zip(frames, anglesZAll) for x in co])
        fc.update()
        fc.convert_to_keyframes(1, frameCount)
        
        fc = a.fcurves.new("location", index=0, action_group="locHorizontal")
        fc.keyframe_points.add(count=len(frames))
        fc.keyframe_points.foreach_set("co", [x for co in zip(frames, locationsXAll) for x in co])
        fc.update()
        fc.convert_to_keyframes(1, frameCount)
        
        fc = a.fcurves.new("rotation_euler", index=2, action_group="degHorizontal")
        fc.keyframe_points.add(count=len(frames))
        fc.keyframe_points.foreach_set("co", [x for co in zip(frames, anglesXAll) for x in co])
        fc.update()
        fc.convert_to_keyframes(1, frameCount) 
        
        ad = cameraOB.animation_data_create()
        ad.action = a


def add_elements(props, context):    
    user_preferences = bpy.context.preferences.addons['hologramcamera'].preferences
    
    
    if props.camera_type == "HPOSTC":
        camera = HPO_Simple_Camera()
    elif props.camera_type == "HPOHCFT":
        camera = HPO_Hybrid_Fixed_Camera()
    elif props.camera_type == "FPHCFT":
        camera = FP_Hybrid_Fixed_Camera()
        
    else:
        print("Error, unknown camera type")
        return

    camera.SetupCameraObject(props, context, user_preferences.xml_dir)

class HOLOGRAMCAMERA_OT_addcamera(Operator):
    bl_idname = "hologramcamera.addcamera"
    bl_label = "Add hologram camera"
    bl_description = "Add hologram camerea"
    bl_options = {'REGISTER', 'UNDO'}
    
    camera_type : EnumProperty(
        items=[('HPOSTC', "HPO Simple", "Horizontal Parallax-Only Simple Translating Camera (HPOSTC)"),
            ("HPOHCFT", "HPO Hybrid Fixed", "Horizontal Parallax-Only Hybrid Camera with Fixed Target (HPOHCFT)"),
            ("FPHCFT", "FP Hybrid Fixed", "Full-Parallax Hybrid Camera with Fixed Target (FPHCFT)")],
        default='FPHCFT',
        name="Type",
        description="Type of camera to add"
        )
    
    size_x : IntProperty(
        name="Width (mm)",
        default=800,
        min=10, soft_min=10,
        max=10000, soft_max=10000,
        description="Width of the hologram"
        )
        
    size_y : IntProperty(
        name="Height (mm)",
        default=600,
        min=10, soft_min=10,
        max=10000, soft_max=10000,
        description="Height of the hologram"
        )
        
    hogel_size : FloatProperty(
        name="Hogel size (mm)",
        default=0.71,
        min=0.1, soft_min=0.1,
        max=10, soft_max=10,
        description="Hogel size"
        )

    number_of_samples : IntProperty(
        name="Number of samples",
        default=512,
        min=1, soft_min=1,
        max=2048, soft_max=2048,
        description="Number of samples to render"
        )
        
    number_of_samples_FP : IntProperty(
        name="Number of samples per axis",
        default=512,
        min=1, soft_min=1,
        max=1024, soft_max=1024,
        description="Number of samples to render per axis (n x n)"
        )    
    
    def draw(self, context):
        layout = self.layout   

        box = layout.box()
        box.prop(self, "camera_type")
             
        box = layout.box()
        box.label(text="Hologram size:")
        box.prop(self, "size_x")
        box.prop(self, "size_y")
        
        box = layout.box()
        box.label(text="Hologram properties:")
        box.prop(self, "hogel_size")     
        
        box = layout.box()
        if (self.camera_type != 'FPHCFT'):
            box.label(text="Number of sampes:")
            box.prop(self, "number_of_samples") 
        else:
            box.label(text="Number of sampes per axis:")
            box.prop(self, "number_of_samples_FP") 
        
    @classmethod
    def poll(cls, context):
        return context.scene is not None

    def execute(self, context):

        # turn off 'Enter Edit Mode'
        use_enter_edit_mode = bpy.context.preferences.edit.use_enter_edit_mode
        bpy.context.preferences.edit.use_enter_edit_mode = False

        add_elements(self, context)

        #restore pre operator state
        bpy.context.preferences.edit.use_enter_edit_mode = use_enter_edit_mode
        return {'FINISHED'}
        


class INFO_MT_hologram(Menu):
    # Define the "Extras" menu
    bl_idname = "VIEW3D_MT_hologram"
    bl_label = "Hologram"

    def draw(self, context):
        layout = self.layout
        layout.operator_context = 'INVOKE_REGION_WIN'
        layout.operator(bpy.types.HOLOGRAMCAMERA_OT_addcamera.bl_idname, icon="RESTRICT_RENDER_OFF")


# Define "Hologram" menu
def menu_func(self, context):
    layout = self.layout
    layout.separator()
    layout.menu(bpy.types.VIEW3D_MT_hologram.bl_idname, text="Hologram", icon="AXIS_SIDE")

        
# Register
classes = [
    HOLOGRAMCAMERA_OT_addcamera,
    INFO_MT_hologram
]

def register_addcamera():
    from bpy.utils import register_class
    for cls in classes:
        register_class(cls)

    bpy.types.VIEW3D_MT_add.append(menu_func)


def unregister_addcamera():
    from bpy.utils import unregister_class
    for cls in reversed(classes):
        unregister_class(cls)

    bpy.types.VIEW3D_MT_add.remove(menu_func)

if __name__ == "__main__":
    register_addcamera()
